/* -*- C++ -*-
 * File:	tiff-fr.cpp
 * Purpose:	TIFF viewer for wxWindows class library
 * Author:	Wolfram Gloger
 * Created:	1996
 * Updated:	
 */

#ifdef __GNUG__
#pragma implementation
#endif

#include <stdarg.h>
#include <stdlib.h>

/* USE_PROTOTYPES is necesssary for early versions of libtiff */
#define USE_PROTOTYPES 1
#include "tiffio.h"

#include "wx.h"

#include "wxtiff.h"

TiffCanvas::TiffCanvas(wxFrame *frame, int x, int y, int w, int h, long style):
wxCanvas(frame, x, y, w, h, style)
{
}

TiffCanvas::~TiffCanvas(void)
{
}

void
TiffCanvas::OnPaint(void)
{
	((TiffFrame *)GetParent())->UpdateCanvas();
}

void
TiffCanvas::OnEvent(wxMouseEvent& event)
{
	TiffFrame *parent = (TiffFrame *)GetParent();

	if(event.RightIsDown())
		((MyFrame *)parent->GetParent())->Delete(parent);
}

void
TiffCanvas::OnScroll(wxCommandEvent& event)
{
	wxCanvas::OnScroll(event);
}

void
TiffCanvas::OnChar(wxKeyEvent& event)
{
	TiffFrame *parent = (TiffFrame *)GetParent();
	char *s;

	switch(event.KeyCode()) {
	case 'q': case 'Q':
		((MyFrame *)parent->GetParent())->Delete(parent);
		break;
	case 'l': case 'L':
		s = wxFileSelector("Load TIFF file", NULL, NULL, NULL, "*.tif");
		if (s) parent->LoadTiff(s);
		break;
	}
}

TiffFrame::TiffFrame(wxFrame *parent, char *title, const char *filename,
					 int x, int y, int w, int h)
	: wxFrame(parent, title, x, y, w, h), file_name(0),
	  cmap(0), img_resx(0), img_resy(0), bm(0), map_size(0), map_r(0), map_g(0), map_b(0)
{
	canvas = new TiffCanvas(this, 0, 0, 50, 50);
	LoadTiff(filename);
	Fit();
	Show(TRUE);
}

TiffFrame::~TiffFrame()
{
	delete canvas;
	if(cmap) delete cmap;
	if(map_r) delete[] map_r;
	if(map_g) delete[] map_g;
	if(map_b) delete[] map_b;
	if(bm) delete[] bm;
	delete[] file_name;
}

void
TiffFrame::SetName(char *name)
{
	MyFrame *parent = (MyFrame *)GetParent();

	if(file_name) {
		parent->Rename(this, name);
		delete[] file_name;
	}
	file_name = copystring(name);
}

int
TiffFrame::LoadTiff(const char *filename)
{
	TIFF *tif;
	unsigned short bitspersample, samplesperpixel, photometric;
	unsigned long x, y, resx, resy;
	int i;
	unsigned word = 0;
	unsigned char *buf, *ptr, *inp;
	unsigned short *rmap, *gmap, *bmap;

	SetName((char *)filename);

	tif = TIFFOpen(filename, "r");
	if (tif == NULL) {
		fprintf(stderr,"unable to read the TIFF image '%s'\n", filename);
		return -1;
	}
	if(TIFFReadDirectory(tif) < 0) {
		fprintf(stderr, "No TIFF directory in %s\n", filename);
		TIFFClose(tif);
		return -2;
	}
	TIFFSetDirectory(tif, 0);
/*	TIFFPrintDirectory(tif, stderr, 0);*/

	TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample);
	if (bitspersample > 8) {
		fprintf(stderr,
				"Can only work with images that have 1 to 8 bits per pixel\n");
		TIFFClose(tif);
		return -3;
	}

	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &resx);
	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &resy);
	TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);

	if(samplesperpixel > 1) {
		fprintf(stderr, "Only 1 sample per pixel allowed.\n");
		TIFFClose(tif);
		return -4;
	}

	if(bm) {
		delete[] bm;
		bm = 0;
	}
	img_resx = resx; img_resy = resy;
	bm = new unsigned char[resx*resy];

	buf = (unsigned char *)malloc(TIFFScanlineSize(tif));
	if(!bm || !buf) {
		fprintf(stderr, "%s: can't allocate memory for buffer\n", filename);
		TIFFClose(tif);
		return -6;
	}

	/* read the image into the bitmap */
	for(ptr=bm, y=0; y<resy; y++) {
		if(TIFFReadScanline(tif, buf, y, 0) < 0) {
			fprintf(stderr,
					"%s: bad data read on line: %ld of TIFF image\n",
					filename, y);
		}
		inp = buf;
		switch (bitspersample) {
		case 1:
			for (x = 0; x < resx; x++) {
				if(x%8 == 0) word = *inp++;
				else word <<= 1;
				*ptr++ = (word & (1u << 7)) ? 0xFFu : 0u;
			}
			break;
		case 2:
			for (x = 0; x < ((resx + 3) / 4); x++) {
				*ptr++ = (*inp >> 6) & 3;
				*ptr++ = (*inp >> 4) & 3;
				*ptr++ = (*inp >> 2) & 3;
				*ptr++ = *inp++ & 3;
			}
			break;
		case 4:
			for (x = 0; x < resx / 2; x++) {
				*ptr++ = *inp >> 4;
				*ptr++ = *inp++ & 0xf;
			}
			break;
		case 8:
			for(x=0; x<resx; x++) *ptr++ = *inp++;
			break;
		default:
			fprintf(stderr, "%s: bad bits/sample: %d\n",
					filename, bitspersample);
			break;
		}
	}
    while(y++ < resy) for(x=0; x<resx; x++) *ptr++ = 0;

	free(buf);

	/* set the colormap */
	if(map_r) delete[] map_r;
	if(map_g) delete[] map_g;
	if(map_b) delete[] map_b;
	map_size = 1 << bitspersample;
	map_r = new unsigned char[map_size];
	map_g = new unsigned char[map_size];
	map_b = new unsigned char[map_size];

	if(TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric)) {
		switch(photometric) {
		case PHOTOMETRIC_MINISBLACK:
			for(i=0; i<map_size; i++)
			 map_r[i] = map_g[i] = map_b[i] = i;
			break;
		case PHOTOMETRIC_MINISWHITE:
			for(i=0; i<map_size; i++)
			 map_r[i] = map_g[i] = map_b[i] = (map_size-1) - i;
			break;
		case PHOTOMETRIC_PALETTE:
			if (TIFFGetField(tif, TIFFTAG_COLORMAP, &rmap, &gmap, &bmap)) {
				/* check if the colormap is 'flat' */
				for(i=0; i<map_size; i++) {
					if(rmap[i]>255 || gmap[i]>255 || bmap[i]>255) break;
				}
				if(i>=map_size) { /* scaling is not necessary */
					for(i=0; i<map_size; i++) {
						/* fill the image colormap structure */
						map_r[i] = rmap[i];
						map_g[i] = gmap[i];
						map_b[i] = bmap[i];
					}
				} else {
					for(i=0; i<map_size; i++) {
						map_r[i] = rmap[i]/256;
						map_g[i] = gmap[i]/256;
						map_b[i] = bmap[i]/256;
					}
				}
			}
			break;
		default:
			printf("Warning: can't handle photometric in '%s'\n", filename);
		}
	}

	/* close the TIFF file */
	TIFFClose(tif);

	canvas->SetSize(0, 0, img_resx, img_resy);
	Fit();
	if(map_size>0 && map_r) {
		if(cmap) delete cmap;
		cmap = new wxColourMap();
		cmap->Create(map_size, map_r, map_g, map_b, TRUE);
		canvas->SetColourMap(cmap);
	}

	return 0;
}

void
TiffFrame::UpdateCanvas()
{
	if (bm) canvas->DrawBitmap8(bm, img_resx, img_resy, 0, 0.0, 0.0);
}

/*
 * Local variables:
 * tab-width: 4
 * End:
 */
